/*!
 * jGoBoard v1.0
 * http://www.jgoboard.com/
 *
 * This software is licensed under a Creative Commons Attribution-NonCommercial 3.0 Unported License:
 * http://creativecommons.org/licenses/by-nc/3.0/
 * If you want to negotiate on a commercial license, please contact the author.
 *
 * Date: 2011-02-17
 */
 
/* jGoBoard relies on jQuery, so make sure it is included before this file. */

// fixed intersection types - DO NOT CHANGE VALUES, they are used as indices in three-item arrays!
var JGO_CLEAR = 0, JGO_BLACK = 1, JGO_WHITE = 2; 

var jgo_coord = "ABCDEFGHJKLMNOPQRST".split(''); // "J18" style of coordinates
var jgo_sgfCoord = "abcdefghijklmnopqrstuvwxyz".split(''); // "ai" style of coordinates

var jgo_errorStr = false;

/**
 * Return last error.
 *
 * @returns Error string or false if no error yet
 */
function jgo_error() {
	return jgo_errorStr;
}

/**
 * Create a helper class to create coordinates from (1,2) (zero-based), "ah" and "J18" types of input.
 * You can create a coordinate with no arguments, in which case it defaults to (0,0), or with one argument,
 * in which case it tries to parse "J18" or "ai" type of string coordinate, or with two arguments, (i,j).
 */
function JGOCoordinate(i, j) {
	if(i != undefined) {
		if(j != undefined) {
			this.i = i;
			this.j = j;
		} else { // try to parse coordinates from first parameter
			this.i = 0;
			this.j = 0;
		
			if(typeof i != "string")
				return;
		
			if(i.substr(0,1).toUpperCase() == i.substr(0,1)) { // capital letter, assume "J18" type
				this.i = $.inArray(i.substr(0,1), jgo_coord); // now also works for IE
				this.j = parseInt(i.substr(1))-1;
			} else { // assume SGF-type coordinate
				this.i = $.inArray(i.substr(0,1), jgo_sgfCoord);
				this.j = $.inArray(i.substr(1), jgo_sgfCoord);
			}
		}
	} else { // called without both parameters
		this.i = 0;
		this.j = 0;
	}
}

/**
 * Make a human readable "J18" type string representation of the coordinate.
 *
 * @returns String representation
 */
JGOCoordinate.prototype.toString = function() {
	return jgo_coord[this.i] + (this.j+1);
}

/**
 * Make an SGF-type "ai" string representation of the coordinate.
 *
 * @returns String representation
 */
JGOCoordinate.prototype.sgfCoord = function() {
	return jgo_sgfCoord[this.i] + jgo_sgfCoord[this.j];
};

/**
 * Make a copy of this coordinate.
 *
 * @returns {JGOCoordinate} A copy of this coordinate
 */
JGOCoordinate.prototype.copy = function() {
	return new JGOCoordinate(this.i, this.j);
};

/**
 * Go board class which has several helper methods to deal with common tasks such
 * as adjacent stones/liberties to a coordinate, or searching connected stones.
 * All changes to board contents need to go through the set(c,s) method!
 *
 * @param {int} width The width of the board
 * @param {int} height The height of the board - if not set, a square board is created
 */
function JGOBoard(width, height) {
	this.width = width;
	
	if(height != undefined)
		this.height = height;
	else
		this.height = this.width;
		
	this.board = new Array();
	
	for(var i=0; i<width; ++i) {
		var column = new Array();
		
		for(var j=0; j<height; ++j)
			column.push(JGO_CLEAR);
		
		this.board.push(column);
	}	
}

/**
 * Get the contents of the board at given coordinate.
 *
 * @param {JGOCoordinate} c The coordinate.
 * @returns {int} Either JGO_CLEAR, JGO_BLACK, or JGO_WHITE
 */
JGOBoard.prototype.get = function(c) {
	return this.board[c.i][c.j];
};
	
/**
 * Get the contents of the board at given coordinate(s).
 * @param c The coordinate as JGOCoordinate object, or an array of JGOCoordinates
 */
JGOBoard.prototype.set = function(c, s) {
	if(c instanceof JGOCoordinate) {
		this.board[c.i][c.j] = s;
	} else if(c instanceof Array) {
		for(var i=0, len=c.length; i<len; ++i)
			this.board[c[i].i][c[i].j] = s;
	}
};
	
/**
 * Get neighboring stones of given type as coordinate array.
 *
 * @param {JGOCoordinate} c The coordinate
 * @param {int} s The type of stones to look for (JGO_CLEAR / JGO_BLACK / JGO_WHITE)
 * @returns {Array} The array of adjacent coordinates of given type (may be an empty array)
 */
JGOBoard.prototype.getAdjacent = function(c, s) {
	var coordinates = [], i = c.i, j = c.j;
				
	if(i>0 && this.board[i-1][j] == s)
		coordinates.push(new JGOCoordinate(i-1, j));
	if(i+1<this.width && this.board[i+1][j] == s)
		coordinates.push(new JGOCoordinate(i+1, j));
	if(j>0 && this.board[i][j-1] == s)
		coordinates.push(new JGOCoordinate(i, j-1));
	if(j+1<this.height && this.board[i][j+1] == s)
		coordinates.push(new JGOCoordinate(i, j+1));

	return coordinates;
};
	
/**
 * Search all stones belonging to the group at the given coordinate - can also be used to find clear areas.
 *
 * @param {JGOCoordinate} cood The coordinate from which to start search - defines also the type of neighbors to search
 * @retuns {Array} A list of connected coordinates of the same type
 */
JGOBoard.prototype.getGroup = function(coord) {
	var color = this.get(coord), // color which we are searching
		added = {}, // hash to check if a coordinate has been added
		queue = [coord], // queue of coordinates to check
		coordinates = [], // group stone coordinates
		board = this; // needs to be saved so it can be accessed inside $.each() iterator
			
	while(queue.length > 0) { // continue until new stones are exhausted
		newQueue = [];
		$.each(queue, function(i,c) {
			if(!(c.toString() in added)) { // a new coordinate found
				coordinates.push(c);
				added[c.toString()] = c;
				newQueue = newQueue.concat(board.getAdjacent(c, color)); // add connected stones to queue
			}
		});
		queue = newQueue;
	}
	
	return coordinates;
};
	
/**
 * Check if an array of coordinates have any liberties.
 *
 * @param {Array} group An array of JGOCoordinate instances comprising the group.
 */
JGOBoard.prototype.hasLiberties = function(group) {
	for(var i=0, len=group.length; i<len; ++i)
		if(this.getAdjacent(group[i], JGO_CLEAR).length > 0)
			return true;
			
	return false;
}

/**
 * Clear the board. 
 *
 * @param {String} coord The board coordinate, such as "K18", "A1"
 */
JGOBoard.prototype.clear = function() {
	var c = new JGOCoordinate(0,0);
	
	for(c.i=0; c.i<this.width; c.i++)
		for(c.j=0; c.j<this.height; c.j++)
			this.set(c, JGO_CLEAR);
}

/**
 * Get the difference between another board setup - i.e. the minimal amount of clear intersections, 
 * black and white stones to add to this board to make it identical to another board.
 *
 * @param {JGOBoard} aBoard The target board
 * @retuns {Array} Array with three lists of coordinates: coordinates to clear, coordinates to fill with black, coordinates to fill with white
 */
JGOBoard.prototype.getDiff = function(aBoard) {
	var c = new JGOCoordinate(0,0), diff = [[],[],[]], a;
	
	for(c.i=0; c.i<19; c.i++)
		for(c.j=0; c.j<19; c.j++)
			if(this.get(c) != (a = aBoard.get(c)))
				diff[a].push(c.copy());
	
	return diff;
}

/**
 * Update the contents of the board using a diff element (create one with getDiff()).
 *
 * @param {Array} diff The diff to execute
 */
JGOBoard.prototype.setDiff = function(diff) {
	this.set(diff[JGO_CLEAR], JGO_CLEAR);
	this.set(diff[JGO_BLACK], JGO_BLACK);
	this.set(diff[JGO_WHITE], JGO_WHITE);
};

/**
 * Update the board to match another board.
 *
 * @param {JGOBoard} b The another board
 */
JGOBoard.prototype.updateBoard = function(b) {
	var c = new JGOCoordinate(0,0);
	
	for(c.i=0; c.i<19; c.i++) {
		for(c.j=0; c.j<19; c.j++) {
			stone = b.board[c.i][c.j];
			
			if(this.board[c.i][c.j] != stone)
				this.set(c, stone);
		}
	}
}

/**
 * Make a copy of this board.
 *
 * @retuns {JGOBoard} A copy of this board (no shared objects)
 */
JGOBoard.prototype.copy = function() {
	var b = new JGOBoard(this.width, this.height);
	
	b.updateBoard(this);
	
	return b;
}

/**
 * Generate a 19x19 go board inside a table element. Basically makes a 21x21 table with inner cells as 
 * intersections - these cells have their coordinates (like "K12", "A4", etc.) as their ids. The 
 * resulting structure is very straightforward and you can inspect it using FireBug if you want, but
 * you don't need to because the method returns a JGOBoard which is tied to the table element and will
 * update all changes to the board in the table.
 *
 * @param {jQuery} board The board table element as jQuery object - created via statement such as $('#board')
 */
function jgo_generateBoard(boardElement) {
	var table = [], i, j;
	
	// create header row, 1+19+1 elements
	table.push('<tr><td class="jgo_top jgo_left">&nbsp;</td>');
	for(i=0; i<19; ++i) {
		table.push('<td class="jgo_top jgo_center">');
		table.push(jgo_coord[i]);
		table.push('</td>');
	}
	table.push('<td class="jgo_top jgo_right">&nbsp;</td></tr>');
	
	for(i=0; i<19; ++i) {
		// create a row with 1 cell for left border, 19 for intersections and 1 for right border
		table.push('<tr><td class="jgo_middle jgo_left">');
		table.push(i+1);
		table.push('</td>');
		for(j=0; j<19; ++j) {
			var coord = jgo_coord[j] + (i+1), shadowClass = "jgo_sh";
			
			if(j==0)
				shadowClass += " jgo_left";
			else if(j==18)
				shadowClass += " jgo_right";

			if(i==0)
				shadowClass += " jgo_top";
			else if(i==18)
				shadowClass += " jgo_bottom";
				
			table.push('<td class="jgo_middle jgo_center" id="');
			table.push(coord);
			table.push('">');
			table.push('<div class="jgo_cont">');
			table.push('<div class="');
			table.push(shadowClass);
			table.push('" id="jgo_sh_');
			table.push(coord);
			table.push('">');
			table.push('</div>');
			table.push('<div class="jgo_st" id="jgo_st_');
			table.push(coord);
			table.push('">');
			table.push('</div>');
			table.push('</div>');
			table.push('</td>');
		}
		table.push('<td class="jgo_middle jgo_right">');
		table.push(i+1);
		table.push('</td></tr>');
	}

	// create footer row, 1+19+1 elements
	table.push('<tr><td class="jgo_bottom jgo_left">&nbsp;</td>');
	for(i=0; i<19; ++i) {
		table.push('<td class="jgo_bottom jgo_center">');
		table.push(jgo_coord[i]);
		table.push('</td>');
	}
	table.push('<td class="jgo_bottom jgo_right">&nbsp;</td></tr>');
	
	boardElement.html(table.join(''));
	
	var board = new JGOBoard(19,19);
	var oldSet = board.set;
	var setBoard = function(c,s) {
		switch(s) {
		case JGO_CLEAR:
			$("#jgo_st_"+c).removeClass("jgo_w").removeClass("jgo_b");
			$("#jgo_sh_"+c).hide();
			break;
		case JGO_BLACK:
			$("#jgo_st_"+c).removeClass("jgo_w").addClass("jgo_b");
			$("#jgo_sh_"+c).show();
			break;
		case JGO_WHITE:
			$("#jgo_st_"+c).removeClass("jgo_b").addClass("jgo_w");
			$("#jgo_sh_"+c).show();
			break;
		}
	};
	
	// override the set method, works on both arrays and individual coordinates
	board.set = function(c,s) {
		if(c instanceof Array)
			$.each(c, function(i, coord) { setBoard(coord, s); });
		else
			setBoard(c, s);
		
		oldSet.call(this, c, s); // also call the old method
	}
	
	board.click = undefined; // this will be called if it is set
	boardElement.click(function(event) {
		var td, id;
		
		if(board.click == undefined) // do nothing if no click listener
			return;
			
		/* The user may have clicked on the TABLE element, TR element, 
		TD element or an element inside a TD element - the following
		code will either exit or give us the TD element */
		if(event.target.tagName == "TD") {
			td = $(event.target)
		} else {
			td = $(event.target).parents("td"); // find all TD parent elements
			if(td.length == 0) // user clicked on TABLE or TR element -> do nothing
				return;		
			td = td.first(); // select the nearest td parent element, in case we have nested tables
		}
		
		if((id = td.attr("id")) != "")
			board.click(new JGOCoordinate(id));  // TD id is a pure coordinate such as "J18"
	});
	
	return board; // return the spiced-up board object
}
